# sciencecloudaccess.py - script functions developed for integration with ScienceCloud via PipelinePilot server
# Use of this code is subject to the terms of the StarDrop License Agreement and use, copying or redistribution of this file for use by other persons or organisations is prohibited
# Copyright(C) Optibrium Ltd 2017
# stardrop-support@optibrium.com
#
"""
Module controls access to the ScienceCloud server from StarDrop via REST calls to a PipelinePilot server.
"""
import PythonQt
import sys
import time
import assaydata
import sciencecloudconfig as config
import credentials
import datasetcreator
import protocolrunner

credentials_manager = None


def _getCredentials():
    """Get a user name and password from the user and store in a CredentialsManager object for use during the lifetime of the session.
    """
    global credentials_manager
    if not credentials_manager or credentials_manager.configured == False:
        if config.USE_SERVICE_ACCOUNT:
            credentials_manager = credentials.CredentialsManager(config.USER, config.PASSWORD, config)
        else:
            dlg = PythonQt.LogInDialog()
            dlg.setWindowTitle("ScienceCloud Login")
            if dlg.run():
                try:
                    credentials_manager = credentials.CredentialsManager(dlg.getUser(), dlg.getPassword(), config)
                except ValueError as ve:
                    print(ve)


def _selectLibrary(stardrop):
    """Retrieve a lost of avaiable libraries from ScienceCouud and allow user to choose one.
    @param: stardrop: The StarDrop application interface.
    @return: Tuple containing the selected library id and a string representing the aggregation type to use (compound or batch)
    """
    try:
        data = _runProtocol(stardrop, protocol="collect_groups_web", credentials=credentials_manager)
        libraries = {}
        if "error" not in data:
            for project, library in zip(data["Projects"], data["Groups"]):
                if project not in libraries:
                    libraries[project] = []
                libraries[project].append(library)
            dlg = PythonQt.ScienceCloudSelection(stardrop.mainWindow())
            dlg.setAvailableItems(libraries)
            if dlg.run():
                result = (dlg.selectedItems(), dlg.aggregationMode())
                dlg.delete()
                return result
        else:
            stardrop.show(data["error"])
    except Exception as e:
        stardrop.show(e.message)
    return "", ""


def _selectQuery(stardrop):
    """Retrieve a list of avaiable queries from ScienceCloud and allow user to choose one.
    @param: stardrop: The StarDrop application interface.
    @return: Tuple containing the selected query id and a string representing the aggregation type to use (compound or batch)
    """
    try:
        data = _runProtocol(stardrop, protocol="collect_queries_web", credentials=credentials_manager)
        queries = {"Saved Queries": []}
        query_lookup = {}
        if "error" not in data:
            for name, query_id in zip(data["query_names"], data["query_ids"]):
                queries["Saved Queries"].append(name)
                query_lookup[name] = query_id
            dlg = PythonQt.ScienceCloudSelection(stardrop.mainWindow())
            dlg.setAvailableItems(queries)
            if dlg.run():
                # TODO: check for multiple queries
                query = dlg.selectedItems()[0]
                mode = dlg.aggregationMode()
                dlg.delete()
                return query_lookup[query], query, mode
        else:
            stardrop.show(data["error"])
    except Exception as e:
        stardrop.show(e.message)
    return "", "", ""


def _getResults(data, mode, unique_assays, entity_data, library):
    """Retrieve and aggregate results from the data returned from the server.
    @param: data: The data returned from the server as headers and rows, where each rows is an indiviidual result
    @param: mode: The aggregation mode to use when collating data
    @param: unique_assays: The list of assays discovered so far, to be added to.
    @param: entity_data: The list of compounds and data discovered so far, to be added to.
    @param: library: The name of the library providing these results.
    @return: Tuple containing the unique assay names, and a map of IDs to assay data
    """
    print("getting results")
    try:
        assay_indexes = []
        result_indexes = {}
        for i, assay_header in enumerate(config.assay_columns):
            try:
                index = data["Headers"].index(assay_header)
            except:
                index = data["Headers"].index(config.raw_assay_columns[i])
            assay_indexes.append(index)
        for i, result_header in enumerate(config.result_columns):
            try:
                index = data["Headers"].index(result_header)
            except:
                index = data["Headers"].index(config.raw_result_columns[i])
            result_indexes[result_header] = index
        # unique_assays = set()
        # entity_data = {}
        for i, row in enumerate(data["Rows"]):
            if config.debug:
                print("Row: %d, data %s" % (i+1, row))
            entries = row.split('\n')
            if config.debug:
                print(len(entries))
            assay = assaydata.createAssayData(entries, assay_indexes)
            unique_assays.add(assay)
            row_data = {r: entries[result_indexes[r]] for r in result_indexes}
            if mode == 'compound':
                entity_id = row_data["Compound ID"]
                secondary_id_name = "Batch ID"
            else:
                entity_id = row_data["Batch ID"]
                secondary_id_name = "Compound ID"
            if entity_id not in entity_data:
                entity_data[entity_id] = {}
                entity_data[entity_id]["secondary_ids"] = set()
                entity_data[entity_id]["libraries"] = set()
                entity_data[entity_id]["compound_tag"] = row_data.get("COMPOUND_TAG", "")
                entity_data[entity_id]["batch_tag"] = []
            if assay not in entity_data[entity_id]:
                entity_data[entity_id][assay] = []
            entity_data[entity_id][assay].append(row_data)
            entity_data[entity_id]["secondary_ids"].add(row_data[secondary_id_name])
            entity_data[entity_id]["libraries"].add(library)
            entity_data[entity_id]["batch_tag"].append(row_data.get("BATCH_TAG", ""))
        return  # unique_assays,entity_data
    except Exception as e:
        print(e)


def _runProtocol(stardrop, protocol, credentials, parameters={}):
    """Execute a PipelinePilot protocol and return the results.
    @param: stardrop: The StarDrop application interface.
    @param: protocol: The protocol name.
    @param: credentials: a CredentialsManager object.
    @param: parameters: Additional parameters to pass to the protocol, as a dictionary.
    @return: The data returned by the protocol.
    """
    try:
        progress = PythonQt.ProgressDialog(stardrop.mainWindow())
        progress.setWindowTitle("ScienceCloud Access")
        progress.modal = True
        interval = 1
        progress.setMessage("Running protocol on ScienceCloud")
        progress.show()
        runner = protocolrunner.ProtocolRunnerThread(protocol, credentials, parameters)
        runner.start()
        while runner.is_alive():
            if progress.cancelled():
                runner.stop()
                runner.join()
                break
            progress.tick()
            time.sleep(interval)
        progress.hide()
        progress.delete()
        return runner.protocolResult()
    except Exception as e:
        try:
            progress.hide()
            print(e)
        except:
            raise RuntimeError("Unable to load ScienceCloud UI. Please check installation")


def _retrieveData(stardrop, library_ids, aggregation_mode, selections=[], silent=False):
    """Retrieve the data for the library selected by the user.
    @param: stardrop: The StarDrop application interface.
    @param: library_ids: A list of library ids to access.
    @param: aggregation_mode: The type of aggregation to use (compound or batch).
    @param: selections: A list of assay identifiers that should be selected when shown to the user (used in refresh, otherwise empty).
    @param: silent: Flag indicating if user should be shown the assay selection dialog again when refreshing
    @return: A StarDrop dataset to be loaded into the UI.
    """
    try:
        assays = set()
        data = {}
        for library_id in library_ids:
            params = {"BatchGroupID": library_id, "AllCompounds": config.all_compounds}
            protocol_output = _runProtocol(stardrop, protocol="assay_results_web", credentials=credentials_manager, parameters=params)
            if "error" not in protocol_output:
                _getResults(protocol_output, aggregation_mode, assays, data, library_id)
            else:
                stardrop.show(protocol_output["error"])
                return None, []
        proceed, selected_assays = assaydata.selectAssays(stardrop, assays, selections, silent)
        if proceed:
            dataset = datasetcreator.createDataset(selected_assays, data, aggregation_mode)
            dataset.addRefreshKey({"function": "ScienceCloud", "library": "\n".join(library_ids), "aggregation": aggregation_mode, "selections": "__A__".join([a.assayID() for a in selected_assays])})
            return dataset, selected_assays
    except Exception as e:
        print(e)


def _retrieveQueryData(stardrop, query_id, aggregation_mode, selections=[], silent=False):
    """Retrieve the data for the library selected by the user.
    @param: stardrop: The StarDrop application interface.
    @param: query_id: The id of the query to access.
    @param: aggregation_mode: The type of aggregation to use (compound or batch).
    @param: selections: A list of assay identifiers that should be selected when shown to the user (used in refresh, otherwise empty).
    @param: silent: Flag indicating if user should be shown the assay selection dialog again when refreshing
    @return: A StarDrop dataset to be loaded into the UI.
    """
    try:
        assays = set()
        data = {}
        params = {"SCQueryID": query_id}
        protocol_output = _runProtocol(stardrop, protocol="query_results_web", credentials=credentials_manager, parameters=params)
        if "error" not in protocol_output:
            _getResults(protocol_output, aggregation_mode, assays, data, query_id)
        else:
            stardrop.show(protocol_output["error"])
            return None, []
        proceed, selected_assays = assaydata.selectAssays(stardrop, assays, selections, silent)
        if proceed:
            dataset = datasetcreator.createDataset(selected_assays, data, aggregation_mode, False)
            dataset.addRefreshKey({"function": "ScienceCloud", "query": query_id, "aggregation": aggregation_mode, "selections": "__A__".join([a.assayID() for a in selected_assays])})
            return dataset, selected_assays
    except Exception as e:
        print(e)


def accessLibrary(stardrop):
    """Respond to a user selecting the ScienceCloud menu item. Obtain a list of libraries for the user to select from,
    and then download the appropriate data to display in StarDrop.
    @param: stardrop: The StarDrop application interface.
    """
    _getCredentials()
    try:
        library_ids, aggregation_mode = _selectLibrary(stardrop)
    except:
        stardrop.show("Unable to load ScienceCloud UI. Please check installation")
    if library_ids and library_ids != []:
        dataset, selected_assays = _retrieveData(stardrop, library_ids, aggregation_mode)
        if dataset:
            dataset_name = "ScienceCloudData"
            if len(library_ids) == 1:
                dataset_name = library_ids[0]
            stardrop.exportDS(dataset_name, dataset.pack())


def accessQuery(stardrop):
    """Respond to a user selecting the ScienceCloud menu item. Obtain a list of queries for the user to select from,
    and then download the appropriate data to display in StarDrop.
    @param: stardrop: The StarDrop application interface.
    """
    _getCredentials()
    try:
        query_id, query_name, aggregation_mode = _selectQuery(stardrop)
    except:
        stardrop.show("Unable to load ScienceCloud UI. Please check installation")
    if query_id:
        dataset, selected_assays = _retrieveQueryData(stardrop, query_id, aggregation_mode)
        if dataset:
            dataset_name = query_name
            stardrop.exportDS(dataset_name, dataset.pack())


try:
    import refreshscript
    appui = sys.modules["__main__"].appui

    def refreshQuery(method, params):
        """Respond to a refresh data command from StarDrop.
        @param: method: The method specified when creating the data set.
        @param: params: The other parameters specified during creation.
        @returns: A python data set as a bute array to be merged with the existing version.
        """
        if method == "ScienceCloud" and ("library" in params or "query" in params):
            _getCredentials()
            try:
                silent = params.get("silent", "False")
                if "library" in params:
                    dataset, selected_assays = _retrieveData(appui, params["library"].split('\n'), params.get("aggregation", "compound"), params.get("selections", "").split("__A__"), silent == "True")
                else:
                    dataset, selected_assays = _retrieveQueryData(appui, params["query"], params.get("aggregation", "compound"), params.get("selections", "").split("__A__"), silent == "True")
                if dataset:
                    return dataset.pack()
            except Exception as e:
                print(e)
        raise RuntimeError("Invalid refresh method")

    refreshscript.setRefreshFunction(refreshQuery)
except:
    pass
